/**
 * Copyright (C) 2022, Protein Bioinformatics Research Group, RCNS
 *
 * Licensed under CC BY-NC 4.0, see LICENSE file for more info.
 *
 * @author Gabor Tusnady <tusnady.gabor@ttk.hu>
 * @author Csongor Gerdan <gerdan.csongor@ttk.hu>
 */

import { StructureElement, Unit } from '../../mol-model/structure';
import { ColorTheme } from '../../mol-theme/color';
import { ThemeDataContext } from '../../mol-theme/theme';
import { Color } from '../../mol-util/color';
import { ColorNames } from '../../mol-util/color/names';
import { ParamDefinition as PD } from '../../mol-util/param-definition';
import { Location } from '../../mol-model/location';
import { ChainList, getResidue, PDBTMChain } from './types';
import { TmDetChainListCache, TmDetDescriptorCache } from './prop';


export type TmDetColorThemeParams = {
    pdbtmDescriptor: any
};

export function TmDetColorTheme(
    ctx: ThemeDataContext,
    props: PD.Values<TmDetColorThemeParams>
): ColorTheme<TmDetColorThemeParams> {
    let descriptorChains: PDBTMChain[] = [];
    let pdbtmDescriptor = props.pdbtmDescriptor;
    const pdbId = ctx.structure?.model.entryId;

    // If it is not given as parameter,
    // but there is one in the cache for this structure.
    if (!pdbtmDescriptor) {
        if (pdbId) {
            pdbtmDescriptor = TmDetDescriptorCache.get(pdbId);
        }
    }
    if (pdbtmDescriptor) {
        descriptorChains = pdbtmDescriptor.chains;
    }

    const chainList = createResidueListsPerChain(descriptorChains);
    if (pdbId) {
        TmDetChainListCache.set(pdbId, chainList);
    }

    return {
        factory: TmDetColorTheme,
        granularity: 'group', //'residue' as any,
        color: (location: Location) => getColor(location, chainList),
        props: props,
        description: 'TMDet...',
    };
}

const DefaultResidueColor = ColorNames.lightgrey;

//TODO: colors of not curated sites
const siteColors = [
    Color.fromArray([255,100,100], 0), // Side1
    Color.fromArray([100,100,255], 0), // Side2
    Color.fromArray([255,255,  0], 0), // TM alpha
    Color.fromArray([255,255,  0], 0), // TM beta
    Color.fromArray([255,127,  0], 0), // TM re-entrant loop
    Color.fromArray([0,255,    0], 0), // Interfacial Helix
    Color.fromArray([196,196,196], 0), // Unknow localization
    Color.fromArray([0,255,    0], 0)  // Membrane Inside
];

const regionColorMapFromCss = new Map();

function getColor(location: Location, chains: ChainList): Color {
    let color = DefaultResidueColor;

    // TODO: How to solve cases when chain operation occurs?
    if (StructureElement.Location.is(location) && Unit.isAtomic(location.unit)) {

        const { chainId, residueId } = getChainAndResidueIds(location);
        color = residueColor(chains, chainId!, residueId!);
    }
    return color;
}

export function getChainAndResidueIds(location: Location) {
    let chainId = undefined;
    let residueId = undefined;

    if (StructureElement.Location.is(location) && Unit.isAtomic(location.unit)) {
        const atomicHierarchy = location.unit.model.atomicHierarchy;
        const residueIdx = StructureElement.residueIndex(location);
        const chainIdx = StructureElement.Location.chainIndex(location);

        residueId = atomicHierarchy.residues.auth_seq_id.value(residueIdx).toString();
        chainId = atomicHierarchy.chains.auth_asym_id.value(chainIdx).toString();
    }
    return {
        chainId: chainId,
        residueId: residueId
    };
}

export function createResidueListsPerChain(chains: PDBTMChain[]) {
    const chainList: ChainList = [];

    chains.forEach((chain: PDBTMChain) => {
        chainList.push({ chainId: chain.chain_label, residues: [] });
        chain.residues.forEach((residue) => {
            chainList[chainList.length - 1].residues.push({
                authId: residue.pdb_res_label,
                siteId: residue.site_data![0].site_id_ref - 1
            });
        });
    });

    return chainList;
}

function residueColor(chains: ChainList, chainId: string, residueId: string): Color {
    let color = DefaultResidueColor;

    const residue = getResidue(chains, chainId, residueId);
    if (residue) {
        color = siteColors[residue.siteId];
    }

    return color;
}

// Provider

export const TmDetColorThemeProvider: ColorTheme.Provider<TmDetColorThemeParams, 'tmdet-custom-color-theme'> = {
    name: 'tmdet-custom-color-theme',
    label: 'TMDet Topology Theme',
    category: 'TMDet',
    factory: TmDetColorTheme,
    getParams: () => ({ pdbtmDescriptor: { isOptional: true } }),
    defaultValues: { pdbtmDescriptor: undefined },
    isApplicable: () => true,
};



// Colors from CSS rules

function loadRegionColorsFromStyleSheets(prefix: string = 'ult_'): void {
    const sheets: CSSStyleSheet[] = Array.from(document.styleSheets);
    sheets.forEach((sheet: CSSStyleSheet) => {
        const rules: CSSRule[] = Array.from(sheet.cssRules);
        rules.forEach((rule: CSSRule) => fetchRule(rule, prefix));
    });
}

export function updateSiteColors(side1: "Inside"|"Outside"|null): void {
    if (!side1) {
        return;
    }
    loadRegionColorsFromStyleSheets();
    if (regionColorMapFromCss.size == 0) {
        console.warn('Cannot read any region-related color rules');
    }
    const inside = regionColorMapFromCss.get("ult_inside");
    const outside = regionColorMapFromCss.get("ult_outside");
    if (side1 == "Inside") {
        siteColors[0] = inside;
        siteColors[1] = outside;
    } else {
        siteColors[0] = outside;
        siteColors[1] = inside;
    }
}

function fetchRule(rule: CSSRule, prefix: string) {
    let styleRule = rule as CSSStyleRule;
    if (styleRule.selectorText?.startsWith('.' + prefix)) {
        const value = styleRule.style.getPropertyValue('fill');
        const color: Color = getStyleColor(value);
        const key = styleRule.selectorText.slice(1);
        regionColorMapFromCss.set(key, color);
    }
}

function getStyleColor(cssColorText: string): Color {
    const values = cssColorText?.match(/\d+/g);
    let intValues = values?.map(value => parseInt(value));
    if (!intValues) {
        intValues = [ 0, 0, 0 ];
    }
    return Color.fromArray(intValues, 0);
}
